'use strict';

var fs   = require('fs')
  , path = require('path')

  , requireRegExp, templatePath, parseDependencies, readFile, readModule
  , readModules, getModuleText, getModulesText
  , modules, local;

requireRegExp = /(?:^\s*|([^\s])(\s*))require\s*\(\s*(?:'([^']+)'|"([^"]+)"|[^)]+)\s*\)/;
templatePath = __dirname + '/template';

parseDependencies = function (text) {
	var match, paths = [];
	while ((match = text.match(requireRegExp))) {
		if (match[2]) {
			if (match[1] === ".") {
				continue;
			}
		} else if (match[1]) {
			continue;
		}
		var path = match[3] || match[4];
		if (!path) {
			throw new TypeError("Not supported require call: '" + match[0] + "'");
		}
		paths.push(path);
		text = text.slice(text.indexOf(match[0]), match[0].length);
	}
	return paths;
};

readFile = function (file, callback) {
	fs.readFile(file, 'utf-8', function (err, text) {
		if (err) {
			callback(err);
			return;
		}
		var deps;
		try {
			deps = parseDependencies(text);
			callback(null, text, deps);
		} catch (e) {
			callback(e);
		}
	});
};

module.exports = function (input, output, callback) {
	var program;
	readFile(input, function (err, text, dependencies) {
		if (err) {
			callback(err);
			return;
		}
		program = text;
		readModules(path.dirname(input), dependencies, modules, [], function (err) {
			if (err) {
				callback(err);
				return;
			}
			fs.readFile(templatePath, 'utf-8', function (err, text) {
				if (err) {
					callback(err);
					return;
				}
				fs.writeFile(output, text.replace('MODULES', getModulesText(modules))
					.replace('PROGRAM', program), callback);
			});
		});
	});
};
